class BaseError(Exception):
    def __init__(self, message):
        self.message = message


class ConfigError(BaseError):
    pass


class BadResponse(BaseError):
    pass


class BadRequest(BaseError):
    pass


class MalformedQuery(BaseError):
    pass


class MalformedBlockAttributes(BaseError):
    pass


class MalformedCreateAttributes(BaseError):
    pass


class MalformedUpdateAttributes(BaseError):
    pass


class MalformedThreshold(BaseError):
    pass


class MissingDomain(BaseError):
    pass


class MalformedDomain(BaseError):
    pass


class MalformedUUID(BaseError):
    pass


class ProtectedDomainError(BaseError):
    pass


class ProtectedEntitynError(BaseError):
    pass


class UnspecificDeleteRequestError(BaseError):
    pass


class UnspecificUpdateRequestError(BaseError):
    pass


class CreateEntityError(BaseError):
    pass


class UpdateEntityError(BaseError):
    pass



import json
from uuid import UUID


class CustomEncoder(json.JSONEncoder):
    """
    Extension of custom encoder that encodes UUIDs.
    """
    def default(self, obj):
        if isinstance(obj, UUID):
            return obj.__str__()
        return json.JSONEncoder.default(self, obj)


def encode(obj):
    """
    Uses CustomEncoder to dump JSON with UUIDs.
    """
    return json.dumps(obj, cls=CustomEncoder)



from .methods import ClientMethods


class Client(ClientMethods):
    def __init__(
        self,
        token,
        service_address,
        domain=None,
        threshold=80,
        scorer='fuzzywuzzy.default_process',
    ):
        self.token = token
        self.service_address = service_address
        self.headers = {
            'Authorization': 'TOKEN {}'.format(self.token),
            'Content-Type': 'application/json',
        }
        self.domain = domain
        self.threshold = threshold
        self.scorer = scorer
        self.client_check()



import os

from setuptools import find_packages, setup

os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name='django-crosswalk-client',
    version='0.0.4',
    packages=find_packages(exclude=('example',)),
    include_package_data=True,
    license='MIT',
    url="https://github.com/The-Politico/django-crosswalk-client",
    description='Client companion to django-crosswalk.',
    classifiers=[
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
    install_requires=[
        'requests',
        'python-slugify',
        'stringcase',
    ]
)



from .client.client_check import ClientCheck
from .client.set_domain import SetDomain
from .client.set_scorer import SetScorer
from .client.set_threshold import SetThreshold

from .domain.create_domain import CreateDomain
from .domain.delete_domain import DeleteDomain
from .domain.get_domain import GetDomain
from .domain.get_domains import GetDomains
from .domain.update_domain import UpdateDomain

from .entity.alias_by_id import AliasById
from .entity.alias_or_create import AliasOrCreate
from .entity.best_match import BestMatch
from .entity.best_match_or_create import BestMatchOrCreate
from .entity.bulk_create import BulkCreate
from .entity.delete_by_id import DeleteById
from .entity.delete_match import DeleteMatch
from .entity.get_entities import GetEntities
from .entity.get_entity import GetEntity
from .entity.supersede_by_id import SupersedeById
from .entity.update_by_id import UpdateById
from .entity.update_match import UpdateMatch


class ClientMethods(
    ClientCheck,
    BulkCreate,
    DeleteDomain,
    CreateDomain,
    BestMatch,
    BestMatchOrCreate,
    SetScorer,
    SetDomain,
    SetThreshold,
    GetDomain,
    GetDomains,
    GetEntities,
    GetEntity,
    DeleteMatch,
    AliasOrCreate,
    DeleteById,
    UpdateById,
    UpdateMatch,
    UpdateDomain,
    AliasById,
    SupersedeById,
):
    pass



from uuid import UUID

from slugify import slugify

from crosswalk_client.exceptions import (MalformedBlockAttributes,
                                         MalformedCreateAttributes,
                                         MalformedDomain, MalformedQuery,
                                         MalformedThreshold,
                                         MalformedUpdateAttributes,
                                         MalformedUUID, MissingDomain)


def validate_query(function):
    def wrapper(*args, **kwargs):
        query = args[1]
        if not isinstance(query, dict):
            raise MalformedQuery("Query should be a dictionary.")
        if len(query.keys()) > 1:
            raise MalformedQuery(
                "Query should only include one key value pair.")
        if not isinstance(list(query.values())[0], str) or \
                not isinstance(list(query.keys())[0], str):
            raise MalformedQuery("Query key and value should be strings.")
        return function(*args, **kwargs)
    return wrapper


def validate_block_attrs(function):
    def wrapper(*args, **kwargs):
        block_attrs = kwargs.get('block_attrs', {}).copy()
        if not isinstance(block_attrs, dict):
            raise MalformedBlockAttributes(
                "Block attributes must be a dictionary.")
        for key in block_attrs:
            if isinstance(block_attrs[key], dict):
                raise MalformedBlockAttributes(
                    "Nested attributes are not allowed in block attributes."
                )
        return function(*args, **kwargs)
    return wrapper


def validate_create_attrs(function):
    def wrapper(*args, **kwargs):
        create_attrs = kwargs.get('create_attrs', {}).copy()
        if not isinstance(create_attrs, dict):
            raise MalformedCreateAttributes(
                "Create attributes must be a dictionary.")
        for key in create_attrs:
            if isinstance(create_attrs[key], dict):
                raise MalformedCreateAttributes(
                    "Nested attributes are not allowed in create attributes."
                )
        return function(*args, **kwargs)
    return wrapper


def validate_update_attrs(function):
    def wrapper(*args, **kwargs):
        update_attrs = kwargs.get('update_attrs', {}).copy()
        if not isinstance(update_attrs, dict):
            raise MalformedUpdateAttributes(
                "Update attributes must be a dictionary.")
        for key in update_attrs:
            if isinstance(update_attrs[key], dict):
                raise MalformedUpdateAttributes(
                    "Nested attributes are not allowed in update attributes."
                )
        return function(*args, **kwargs)
    return wrapper


def validate_domain(function):
    def wrapper(*args, **kwargs):
        domain = kwargs.get('domain', args[0].domain)
        if domain is None:
            raise MissingDomain("Must set a domain.")
        if not isinstance(domain, str):
            raise MalformedDomain("Domain must be a slug string.")
        if slugify(domain) != domain:
            raise MalformedDomain("Domain should be a slug.")
        return function(*args, **kwargs)
    return wrapper


def validate_threshold(function):
    def wrapper(*args, **kwargs):
        threshold = kwargs.get('threshold', args[0].threshold)
        if not isinstance(threshold, int):
            raise MalformedThreshold("Threshold should be an integer.")
        if threshold < 0 or threshold > 100:
            raise MalformedThreshold("Threshold should be between 0 and 100.")
        return function(*args, **kwargs)
    return wrapper


def validate_uuid(function):
    def wrapper(*args, **kwargs):
        uuid = args[1]
        if not isinstance(uuid, UUID):
            raise MalformedUUID("Invalid UUID.")
        return function(*args, **kwargs)
    return wrapper


def validate_target_uuid(function):
    """
    Target UUIDs are used to set foreign keys and should be either a valid UUID
    or None to unset the foreign key.
    """
    def wrapper(*args, **kwargs):
        uuid = args[2]
        if uuid is not None and not isinstance(uuid, UUID):
            raise MalformedUUID("Invalid UUID for target.")
        return function(*args, **kwargs)
    return wrapper
